
import os
from math import *
import numpy
import numpy as np
import pandas
import pandas as pd
import random
import subprocess
import pytest
import copy
import warnings
from tramway.core.exceptions import *
from tramway.core.scaler import unitrange
from tramway.helper import *
from tramway.helper.roi import *
from tramway.helper.simulation import *
from tramway.analyzer import *

data_version = '200803'

seed = 4294947105


def reset_random_generator(seed):
    if seed is not None:
        random.seed(seed)
        numpy.random.seed(seed)
def gen_spt_data(generator, dir, filename, seed=None):
    filepath = '/'.join((dir, filename))
    if not os.path.isfile(filepath):
        reset_random_generator(seed)
        traj = generator()
        traj.to_csv(filepath, sep='\t', index=False)

def gen_linear_gradient_D():
    return random_walk_2d(
            n_trajs=400,
            N_mean=5,
            dt=.05,
            D0=.1,
            amplitude_D=.05,
            mode_D='linear',
            sigma_noise=1e-4,
            )
def gen_linear_gradient_V():
    return random_walk_2d(
            n_trajs=400,
            N_mean=5,
            dt=.05,
            D0=.1,
            amplitude_V=2.,
            mode_V='potential_linear',
            sigma_noise=1e-4,
            )
def gen_radial_sink_D():
    return random_walk_2d(
            n_trajs=400,
            N_mean=5,
            dt=.05,
            D0=.1,
            amplitude_D=.05,
            mode_D='gaussian_trap',
            sigma_noise=1e-4,
            )
def gen_radial_sink_V():
    return random_walk_2d(
            n_trajs=400,
            N_mean=5,
            dt=.05,
            D0=.1,
            amplitude_V=2.,
            mode_V='potential_force',
            sigma_noise=1e-4,
            )

def gen_moving_sink_V():
    def force(xy, t):
        sink_r = .17
        t1 = 50.
        t2 = 90.
        if t<t1:
            sink_center = ((t1-t)/t1) * np.r_[-.3,-.1] + (t/t1) * np.r_[.1,.1]
        else:
            sink_center = ((t2-t)/(t2-t1)) * np.r_[.1,.1] + ((t-t1)/(t2-t1)) * np.r_[.25,-.45]
        r = xy - sink_center
        norm = np.sqrt(np.dot(r,r))
        if norm < sink_r:
            scaled_gradV = (r/norm) * (1. + cos(norm/sink_r*pi)) / 2.
            f = 5. * (-scaled_gradV)
        else:
            f = np.zeros_like(r)
        return f
    return random_walk(
            diffusivity=.3,
            force=force,
            trajectory_mean_count=10,
            duration=90.,
            box=(-.5,-.5,1.,1.),
            )


@pytest.fixture
def sptdatafiles(tmpdir):
    testdir = os.path.dirname(__file__)
    testdir = '{}/test_analyzer_{}'.format(testdir, data_version)
    if not os.path.isdir(testdir):
        os.makedirs(testdir)
    sptfile0 = 'test00_linear_diffusivity_gradient.txt'
    gen_spt_data(gen_linear_gradient_D, testdir, sptfile0, seed)
    sptfile1 = 'test01_linear_potential_gradient.txt'
    gen_spt_data(gen_linear_gradient_V, testdir, sptfile1, seed)
    sptfile2 = 'test02_radial_diffusivity_sink.txt'
    gen_spt_data(gen_radial_sink_D, testdir, sptfile2, seed)
    sptfile3 = 'test03_radial_potential_sink.txt'
    gen_spt_data(gen_radial_sink_V, testdir, sptfile3, seed)
    filenames = [sptfile0, sptfile1, sptfile2, sptfile3]
    return testdir, filenames

@pytest.fixture
def one_sptdatafile(sptdatafiles):
    testdir, filenames = sptdatafiles
    return '/'.join((testdir, filenames[0]))
@pytest.fixture
def all_sptdatafiles(sptdatafiles):
    testdir, _ = sptdatafiles
    return testdir+'/*.txt'

def timefree(tree):
    if isinstance(tree.data, Maps):
        try:
            tree.data.runtime = None
        except AttributeError:
            pass
    new_tree = type(tree)(tree.data)
    new_tree.metadata = { k:tree.metadata[k]
            for k in tree.metadata
            if k != 'datetime' }
    for label in tree.labels:
        new_tree.add(timefree(tree[label]), label, tree.comments[label])
    return new_tree


class Common(object):

    def equal(self, tree_a, tree_b, label=None):
        try:
            tree_a = tree_a.statefree()
        except AttributeError:
            pass
        if isinstance(tree_a.data, pandas.DataFrame):
            tree_a.data.localization_error = None
            tree_a.data.frame_interval = None
        assert type(tree_a) is type(tree_b)
        if label is None:
            label = self.label
        else:
            label = '_'.join((self.label, label))
        file_a = os.path.join(self.tmpdir, label+'_a.rwa')
        file_b = os.path.join(self.tmpdir, label+'_b.rwa')
        save_rwa(file_a, timefree(tree_a), force=True)
        save_rwa(file_b, timefree(tree_b), force=True)
        h5diff = 'h5diff'
        if os.name == 'nt':
            h5diff += '.exe'
        p = subprocess.Popen((h5diff, file_a, file_b),
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p.communicate()
        if err:
            print(err)
        if out:
            print(out)
        return not (bool(err) or bool(out))

    @property
    def tmpdir(self):
        return self._tmpdir
    @tmpdir.setter
    def tmpdir(self, tmpdir):
        try:
            tmpdir = tmpdir.strpath
        except AttributeError:
            pass
        self._tmpdir = tmpdir


class TestSptDataAccess(Common):

    def test_prms(self, all_sptdatafiles):
        a=RWAnalyzer()
        a.spt_data.from_ascii_files(all_sptdatafiles)
        a.spt_data.localization_precision = .03
        for f in a.spt_data:
            assert f.localization_precision == .03
        for i, f in enumerate(a.spt_data):
            f.localization_precision = float(i) * .01
        assert 1 < i
        try:
            a.spt_data.localization_error
        except AttributeError:
            pass
        else:
            assert False # localization_error is unique and should not
        for f in a.spt_data:
            f.localization_error = .0009
        assert a.spt_data.localization_precision == .03

class TestTesseller(Common):

    def test_full_fov_grid(self, one_sptdatafile, tmpdir):
        self.tmpdir = tmpdir
        self.label = 'full_fov_grid'
        #
        a=RWAnalyzer()
        a.spt_data.from_ascii_file(one_sptdatafile)
        a.spt_data.localization_precision = 1e-4
        a.tesseller.from_callable(tessellers.Squares)
        for i, r in enumerate(a.roi.as_support_regions()):
            assert i == 0
            df = r.discard_static_trajectories(r.crop())
            sampling = a.sampler.sample(df)
            r.add_sampling(sampling, 'full fov grid')
            tree_a = r._spt_data.analyses
        #
        data_b = load_xyt(one_sptdatafile)
        tree_b = Analyses(discard_static_trajectories(data_b, 1e-4**2))
        tessellate(tree_b, 'grid', label='full fov grid')
        tree_b.data = data_b
        #
        assert self.equal(tree_a, tree_b)

    def test_single_roi_kmeans(self, one_sptdatafile, tmpdir):
        self.tmpdir = tmpdir
        self.label = 'single_roi_kmeans'
        #
        a=RWAnalyzer()
        a.spt_data.from_ascii_file(one_sptdatafile)
        a.spt_data.localization_precision = 1e-4
        a.tesseller.from_callable(tessellers.KMeans)
        a.roi.from_squares(numpy.zeros((1,2)), 1.)
        a.sampler.from_voronoi()
        a.sampler.min_location_count = 10
        for i, r in enumerate(a.roi.as_support_regions()):
            assert i == 0
            df = r.discard_static_trajectories(r.crop())
            sampling = a.sampler.sample(df)
            r.add_sampling(sampling, 'single roi kmeans')
            tree_a = r._spt_data.analyses
        #
        data_b = load_xyt(one_sptdatafile)
        tree_b = Analyses(discard_static_trajectories(crop(data_b, [-.5,-.5,.5,.5]), 1e-4**2))
        tessellate(tree_b, 'kmeans', min_n=10, prune=False, label='single roi kmeans')
        tree_b.data = data_b
        #
        assert self.equal(tree_a, tree_b)

    def test_3roi_hexagons(self, one_sptdatafile, tmpdir):
        self.tmpdir = tmpdir
        self.label = '3roi_hexagons'
        #
        roi_bounds = [([-.4,-.5],[0.,-.1]),([-.1,0.],[.3,.4]),([.1,-.3],[.5,.1])]
        roi_bounds = [ (numpy.array(lb), numpy.array(ub)) for lb, ub in roi_bounds ]
        #
        a=RWAnalyzer()
        a.spt_data.from_ascii_file(one_sptdatafile)
        a.tesseller.from_callable(tessellers.Hexagons)
        a.tesseller.ref_distance = .01
        a.tesseller.avg_location_count = 5
        a.roi.from_bounding_boxes(roi_bounds, label='3-roi hexagons', group_overlapping_roi=True)
        b = RoiHelper(one_sptdatafile, verbose=False)
        b.set_bounding_boxes(roi_bounds, collection_label='3-roi hexagons')
        b.tessellate('hexagon', ref_distance=.01, avg_location_count=5)
        for i, r in enumerate(a.roi.as_support_regions()):
            df = r.crop()
            a.tesseller.reset() # let the calibration parameters vary from a roi to another
            sampling = a.sampler.sample(df)
            r.add_sampling(sampling)
            tree_a = r.get_sampling()._subtree
            tree_b = b.analyses[r.label].statefree()
            assert self.equal(tree_a, tree_b)

    def test_dir_gwr(self, all_sptdatafiles, tmpdir):
        self.tmpdir = tmpdir
        self.label = 'dir_gwr'
        #
        a=RWAnalyzer()
        a.spt_data.from_ascii_files(all_sptdatafiles)
        a.tesseller.from_callable(tessellers.GWR)
        a.tesseller.topology = 'approximate density'
        a.tesseller.min_distance = None
        del a.tesseller._explicit_kwargs['min_distance']
        for f in a.spt_data:
            print(f.source)
            #
            df = f.dataframe
            a.tesseller.reset()
            reset_random_generator(seed)
            sampling = a.sampler.sample(df)
            f.add_sampling(sampling, 'dir gwr')
            tree_a = f.analyses
            #
            tree_b = Analyses(load_xyt(f.source))
            reset_random_generator(seed)
            tessellate(tree_b, 'gwr', lifetime=.05, complete_delaunay=True,
                    label='dir gwr')
            #
            #print(sampling)
            #print(tree_b['dir gwr'].data)
            assert self.equal(tree_a, tree_b)


@pytest.fixture
def staticmesh(sptdatafiles):
    testdir, filenames = sptdatafiles
    sptdatafile = '/'.join((testdir, filenames[3]))
    a=RWAnalyzer()
    a.spt_data.from_ascii_file(sptdatafile)
    a.tesseller.from_callable(tessellers.Hexagons)
    f = next(iter(a.spt_data))
    mesh = a.tesseller.tessellate(f.dataframe)
    mesh = Analysis(mesh, f)
    sampling = a.sampler.sample(f.dataframe, mesh)
    sampling.commit_as_analysis('hexagons')
    return f.analyses

@pytest.fixture
def dynamicmesh(tmpdir):
    testdir = os.path.dirname(__file__)
    testdir = '{}/test_analyzer_{}'.format(testdir, data_version)
    if not os.path.isdir(testdir):
        os.makedirs(testdir)
    sptfile = 'test04_moving_potential_sink.txt'
    gen_spt_data(gen_moving_sink_V, testdir, sptfile, seed)
    return '/'.join((testdir, sptfile))


class TestMapper(Common):

    def test_stochastic_dv(self, tmpdir, staticmesh):
        self.tmpdir = tmpdir
        self.label = 'sdv'
        #
        a=RWAnalyzer()
        a.spt_data.from_analysis_tree(copy.deepcopy(staticmesh))
        a.spt_data.localization_precision = 1e-4
        a.mapper.from_plugin('stochastic.dv')
        a.mapper.diffusivity_prior = 0
        a.mapper.potential_prior = 1
        a.mapper.ftol = 1e-3
        a.mapper.worker_count = 1
        a.mapper.verbose = False
        #
        f = next(iter(a.spt_data))
        sampling = f.get_sampling('hexagons')
        reset_random_generator(seed)
        mapping = a.mapper.infer(sampling)
        mapping.commit_as_analysis('sdv')
        tree_a = f.analyses
        #
        tree_b = copy.deepcopy(staticmesh)
        reset_random_generator(seed)
        infer(tree_b, 'stochastic.dv',
                sigma = 1e-4,
                diffusivity_prior = 0,
                potential_prior = 1,
                ftol = 1e-3,
                worker_count = 1,
                input_label = 'hexagons',
                output_label = 'sdv',
                )
        #
        print(tree_a)
        print(tree_b)
        print(tree_a['hexagons']['sdv'].data)
        print(tree_b['hexagons']['sdv'].data)
        assert self.equal(tree_a, tree_b.statefree())

    def test_time_regul(self, dynamicmesh, tmpdir):
        self.tmpdir = tmpdir
        self.label = 'time_regul'
        #
        a=RWAnalyzer()
        a.spt_data.from_ascii_file(dynamicmesh)
        a.spt_data.localization_precision = 1e-4
        a.tesseller.from_plugin('hexagon')
        a.time.from_sliding_window(duration=30)
        a.sampler.from_nearest_neighbors(10)
        a.mapper.from_plugin('stochastic.dv')
        a.mapper.diffusivity_prior = 0
        a.mapper.potential_prior = 1
        a.mapper.potential_time_prior = 1
        a.mapper.ftol = 1e-2
        a.mapper.worker_count = 1
        a.mapper.verbose = False
        #
        f = next(iter(a.spt_data))
        sampling = a.sampler.sample(f.dataframe)
        sampling = commit_as_analysis('hexagons', sampling, f)
        reset_random_generator(seed)
        mapping = a.mapper.infer(sampling)
        mapping.commit_as_analysis('sdv')
        tree_a = f.analyses
        #
        tree_b = Analyses(load_xyt(dynamicmesh))
        tessellate(tree_b, 'hexagon', label='hexagons',
                time_window_duration=30,
                knn=10,
                enable_time_regularization=True)
        reset_random_generator(seed)
        infer(tree_b, 'stochastic.dv',
                sigma=1e-4,
                diffusivity_prior=0,
                potential_prior=1,
                potential_time_prior=1,
                ftol=1e-2,
                worker_count=1,
                output_label='sdv',
                )
        #
        assert self.equal(tree_a, tree_b)


class TestIndexer(Common):

    def test_nooverlap_roi(self, dynamicmesh):
        #
        a=RWAnalyzer()
        a.spt_data.from_ascii_file(dynamicmesh)
        a.spt_data.localization_precision = 1e-4
        roi_centers = [[.2,-.1],[-.3,.3],[0.,.1],[-.2,-0.]]
        a.roi.from_squares(np.array(roi_centers), .2, group_overlapping_roi=False)
        #
        for i, r in a.roi.as_individual_roi(index=1, return_index=True):
            assert i == 1
            lb,ub = r.bounding_box
            assert np.all( lb == np.array(roi_centers[1])-.1 )
            assert np.all( ub == np.array(roi_centers[1])+.1 )
        #
        j = 3
        for i, r in a.roi.as_individual_roi(index=[3,2,1,0], return_index=True):
            assert i == j
            j -= 1
            lb,ub = r.bounding_box
            assert np.all( lb == np.array(roi_centers[i])-.1 )
            assert np.all( ub == np.array(roi_centers[i])+.1 )
        #
        j = 0
        for r in a.roi.as_individual_roi(index=set((0,3))):
            lb,ub = r.bounding_box
            assert np.all( lb == np.array(roi_centers[j])-.1 )
            assert np.all( ub == np.array(roi_centers[j])+.1 )
            j += 3
        #
        for r0, r1 in zip(
                a.roi.as_individual_roi(),
                a.roi.as_support_regions(),
                ):
            lb0,ub0 = r0.bounding_box
            lb1,ub1 = r1.bounding_box
            assert np.all( lb0 == lb1 )
            assert np.all( ub0 == ub1 )

    def test_overlapping_roi(self, dynamicmesh):
        #
        a=RWAnalyzer()
        a.spt_data.from_ascii_file(dynamicmesh)
        a.spt_data.localization_precision = 1e-4
        roi_centers = [[.2,-.1],[-.3,.3],[0.,.1],[-.2,-0.]]
        a.roi.from_squares(np.array(roi_centers), .2, group_overlapping_roi=True)
        #
        j = 0
        for i,r in a.roi.as_individual_roi(return_index=True):
            assert i == j
            lb,ub = r.bounding_box
            assert np.all( lb == np.array(roi_centers[j])-.1 )
            assert np.all( ub == np.array(roi_centers[j])+.1 )
            j += 1
        #
        for i,r in a.roi.as_support_regions(return_index=True):
            if i==0:
                lb,ub = r.bounding_box
                assert np.isclose(lb[0], -.3)
                assert np.isclose(lb[1], -.2)
                assert np.isclose(ub[0], .3)
                assert np.isclose(ub[1], .2)
            elif i==1:
                lb,ub = r.bounding_box
                assert lb[0] == roi_centers[1][0]-.1
                assert lb[1] == roi_centers[1][1]-.1
                assert ub[0] == roi_centers[1][0]+.1
                assert ub[1] == roi_centers[1][1]+.1
            else:
                assert i<2 # False

    def test_time_segments(self, dynamicmesh):
        #
        a=RWAnalyzer()
        a.spt_data.from_ascii_file(dynamicmesh)
        #roi = [[.2,-.1],[-.3,.3],[0.,.1],[-.2,-0.]]
        #a.roi.from_squares(np.array(roi), .2)
        #
        a.time.from_sliding_window(duration=10)
        a.time.sync_start_times()
        #
        for r in a.roi.as_support_regions(index=0):
            df = r.crop()
            sampling = a.sampler.sample(df)
            for t, s in a.time.as_time_segments(sampling):
                assert t[0]<=s.points['t'].min()
                assert s.points['t'].max()<=t[1]
        #
        a.mapper.from_plugin('d')
        j = 3
        for d in a.spt_data:
            for r in d.roi.as_support_regions(index=0):
                df = r.crop()
                sampling = a.sampler.sample(df)
                for i,t,s in a.time.as_time_segments(sampling, index=j, return_index=True):
                    assert i == j
                    assert np.all(t == sampling.tessellation.time_lattice[j])
                    assert t[0]<=s.points['t'].min()
                    assert s.points['t'].max()<=t[1]


class TestAssignment(object):
    def test_readonly_properties(self):
        a = RWAnalyzer()
        for prop in ('pipeline',):
            try:
                setattr(a, prop, True)
            except AttributeError:
                pass
            else:
                assert prop is not prop # the mentioned property is writable
    def test_warnings_for_misplaced_attributes(self, one_sptdatafile):
        warnings.simplefilter('error', MisplacedAttributeWarning)
        a = RWAnalyzer()
        a.spt_data.from_ascii_file(one_sptdatafile)
        for attr, val in (
                ('localization_error', 0),
                ('localization_precision', 0),
                ('columns', list('nxyt')),
                ('scaler', unitrange()),
                ('resolution', .1),
                ('dt', .04),
                ('time_step', .05),
                ('frame_interval', .06),
                ):
            try:
                setattr(a, attr, val)
            except MisplacedAttributeWarning:
                pass
            else:
                assert attr is not attr # no warning on setting the mentioned attribute
    def test_init_for_misplaced_attributes(self):
        warnings.simplefilter('ignore', MisplacedAttributeWarning)
        a = RWAnalyzer()
        for attr, val, permissive in (
                ('localization_error', 0, False),
                ('localization_precision', 0, False),
                ('columns', list('nxyt'), False),
                ('scaler', unitrange(), False),
                ('resolution', .1, False),
                ('dt', .04, False),
                ('time_step', .05, False),
                ('frame_interval', .06, False),
                ):
            try:
                setattr(a, attr, val)
            except AttributeError:
                if permissive:
                    raise # could not set the attribute before initializing its parent
                else:
                    pass
            else:
                if permissive:
                    pass
                else:
                    assert attr is not attr # could set the attribute before initializing its parent


class TestSideEffects(object):
    def test_attribute_overwrite(self):
        warnings.simplefilter('ignore', SideEffectWarning)
        a = RWAnalyzer()
        a.spt_data.from_dataframe(pandas.DataFrame([], columns=list('xyt')))
        a.spt_data.localization_precision = .03
        a.spt_data.frame_interval = .05
        a.time.frame_interval = .04
        old_spt_data = a.spt_data
        a.spt_data = spt_data.from_dataframe(pandas.DataFrame([], columns=list('xyt')))
        a.spt_data.localization_error = old_spt_data.localization_error
        a.spt_data.dt = old_spt_data.dt
        assert a.spt_data.localization_precision == .03
        assert a.spt_data.dt == .04
        warnings.simplefilter('error', SideEffectWarning)


class TestMisc(object):

    def test_analysis_save(self, tmp_path):
        existing_tree = Analyses(0)
        existing_tree['child 0'] = 1
        existing_tree['child 0']['grand child 00'] = 2
        existing_tree['child 1'] = 3
        existing_tree['child 1']['grand child 10'] = 4
        existing_tree['child 2'] = 5
        existing_tree['child 2']['grand child 20'] = 6
        existing_tree['child 2']['grand child 20']['grand grand child 200'] = 10
        _path = tmp_path / 'tree.rwa'
        Analysis.save(_path, existing_tree, force=True)
        #
        extra_tree = Analyses(0)
        extra_tree['child 0'] = 1
        extra_tree['child 0']['grand child 00'] = 7
        extra_tree['child 0']['grand child 01'] = 8
        extra_tree['child 1'] = 3
        extra_tree['child 1']['grand child 11'] = 9
        extra_tree['child 2'] = 5
        Analysis.save(_path, extra_tree, force=True)
        #
        resulting_tree = load_rwa(_path)
        print(resulting_tree)
        assert resulting_tree.data == 0
        assert resulting_tree['child 0'].data == 1
        assert resulting_tree['child 0']['grand child 00'].data == 7
        assert resulting_tree['child 0']['grand child 01'].data == 8
        assert resulting_tree['child 1'].data == 3
        assert resulting_tree['child 1']['grand child 10'].data == 4
        assert resulting_tree['child 1']['grand child 11'].data == 9
        assert resulting_tree['child 2'].data == 5
        assert resulting_tree['child 2']['grand child 20'].data == 6
        assert resulting_tree['child 2']['grand child 20']['grand grand child 200'].data == 10





from .collections import Collections
import scipy.sparse as sparse
import numpy as np

"""
This module exports functions for the identification of regions of interest.
"""


def epanechnikov_density(xy, eval_at, target_pattern_size):
    from sklearn.neighbors import KernelDensity

    bandwidth = .5 * target_pattern_size # the bandwidth is similar to a radius (not a diameter)
    estimator = KernelDensity(kernel='epanechnikov', bandwidth=bandwidth)

    estimator.fit(xy)
    log_density = estimator.score_samples(eval_at)

    return log_density


def density_based_roi(locations, min_kernel_density=.1,
        target_pattern_size=.3, step_size_factor=.1, dr=4,
        kernel_density=epanechnikov_density):
    xy = locations[['x','y']].values
    
    extent = np.ravel(np.r_[xy.min(axis=0, keepdims=True), xy.max(axis=0, keepdims=True)])

    step = step_size_factor * target_pattern_size

    if extent is None:
        xymin, xymax = xy.min(axis=0), xy.max(axis=0)
    else:
        xymin, xymax = [extent[0], extent[1]], [extent[2], extent[3]]

    grid_x, grid_y = np.arange(xymin[0], xymax[0], step), np.arange(xymin[1], xymax[1], step)
    _x, _y = np.meshgrid(grid_x, grid_y, indexing='ij')
    grid = np.c_[_x.reshape((-1,1)),_y.reshape((-1,1))]

    log_density = epanechnikov_density(xy, grid, target_pattern_size)
    log_density = log_density.reshape(len(grid_x),len(grid_y))
    
     # find local maxima using logical convolution
    sizx, sizy = log_density.shape[0]-2*dr, log_density.shape[1]-2*dr
    _x = lambda x: (log_density.shape[0] if x==0 else x)
    _y = lambda y: (log_density.shape[1] if y==0 else y)
    mask = np.ones((sizx, sizy), dtype=bool)
    for dx in range(-dr,dr):
        for dy in range(-dr,dr):
            if (not (dx==0 and dy==0)) and (dx+dy <= dr):
                mask = mask & \
                        (log_density[dr+dx:_x(dx-dr),dr+dy:_y(dy-dr)] < log_density[dr:-dr,dr:-dr])

    local_max = sparse.coo_matrix(mask)
    selected_nz = np.log(min_kernel_density) <= log_density[local_max.row,local_max.col]
    if not np.any(selected_nz):
        print('global max density is below the threshold: {} < {}'.format(
            np.exp(log_density[local_max.row,local_max.col].max()), min_kernel_density))
    selected_i, selected_j = local_max.row[selected_nz], local_max.col[selected_nz]

    # define the roi as center points
    roi_centers = [ np.r_[grid_x[dr+i],grid_y[dr+j]] for i,j in zip(selected_i, selected_j) ]
    if roi_centers:
        if len(roi_centers)<10:
            print('few roi found: {}'.format(len(roi_centers)))
    else:
        raise RuntimeError('no roi found')
        
    return roi_centers


def set_contiguous_time_support_by_count(points, space_bounds, time_window, min_points,
        min_segments=1, group_overlapping_roi=False, start_stop_min_points=None):
    """
    Arguments:

        points (DataFrame): location or translocation data

        space_bounds (dict or sequence): (collections of) boundaries,
            each boundary being a pair of NumPy arrays (lower, upper)

        time_window (float or pair of floats): window duration or
            window duration and shift

        min_points (int): desired minimum number of data rows per time segment

        min_segments (int): desired minimum number of selected time segments.

        group_overlapping_roi (bool): see :class:`~tramway.helper.roi.RoiCollections`

    Returns:

        dict or list: similar to `points`, with an additional column for time
            in each bound array.

    """
    noname = not isinstance(space_bounds, dict)
    if noname:
        space_bounds = {'': tuple(space_bounds)}

    roi = Collections(group_overlapping_roi)
    for label in space_bounds:
        roi[label] = space_bounds[label]

    try:
        duration, shift = time_window
    except (ValueError, TypeError):
        duration = shift = time_window
    min_duration = duration + (min_segments - 1) * shift

    if start_stop_min_points is None:
        start_stop_min_points = min_points

    n_space_dims = sum([ c in points.columns for c in 'xyz' ])

    regions = { label: [] for label in roi }
    for r in roi.regions:
        units = roi.regions.region_to_units(r)

        region_weight = sum([ len(u) for u in units.values() ]) # TODO: relative surface area instead
        if not group_overlapping_roi:
            assert region_weight == 1
        threshold = start_stop_min_points * region_weight

        times = roi.regions.crop(r, points)['t'].values
        if min_points and len(times) < min_points * region_weight * min_segments:
            continue

        counts = np.zeros(times.size, dtype=np.int)
        for t in range(times.size):
            counts[t] = np.sum(times[t]-duration<=times[max(0,t-threshold):t+1])

        ok = np.flatnonzero(threshold<=counts)
        if ok.size==0:
            continue

        first_t, last_t = ok[0], ok[-1]

        if times[last_t] - times[first_t] + duration < min_duration:
            continue

        if min_points:
            if min_points != start_stop_min_points:
                threshold = min_points * region_weight

                ok = threshold<=counts[first_t:last_t+1]

                assert ok[0] and ok[-1]
                if not np.all(ok):
                    ok_first = first_t + 1 + np.flatnonzero(~ok[:-1] & ok[1:])
                    ok_last = first_t + np.flatnonzero(ok[:1] & ~ok[1:])
                    gap = times[ok_first] - times[ok_last]
                    assert np.all(0 < gap)
                    max_gap = window - 2 * shift # TODO: make it an input argument
                    gap_ok = gap <= max_gap
                    if not np.all(gap_ok):
                        first_t = np.r_[first_t,  ok_first[~gap_ok]]
                        last_t = np.r_[ok_last[~gap_ok], last_t]
                        segment_dur = times[ok_last] - times[ok_first] # + window
                        assert np.all(0 <= segment_dur)
                        segment_ok = min_duration <= segment_dur
                        if not np.all(segment_ok):
                            first_t = first_t[~segment_ok]
                            last_t = last_t[~segment_ok]

        if np.isscalar(first_t):
            first_ts, last_ts = [first_t], [last_t]
        else:
            first_ts, last_ts = first_t, last_t

        start_times, end_times = [], []
        for first_t, last_t in zip(first_ts, last_ts):

            min_start_time = times[first_t] - duration
            max_start_time = max(times[0], min_start_time)

            min_end_time = times[last_t]
            if last_t+1 == times.size:
                max_end_time = times[-max(1, threshold)] + duration
            else:
                max_end_time = min_end_time

            max_total_duration = max_end_time - min_start_time
            nsegments = np.floor((max_total_duration - duration) / shift) + 1
            total_duration = duration + (nsegments - 1) * shift
            time_margin = .5 * (max_total_duration - total_duration)
            start_time = min(min_start_time + time_margin, max_start_time)
            end_time = max_end_time# - time_margin # no need to discard the trailing data

            if nsegments<min_segments:
                continue

            start_times.append(start_time)
            end_times.append(end_time)

        for label in units:
            bounds = space_bounds[label]
            new_bounds = regions[label]
            for i in units[label]:
                for start_time, end_time in zip(start_times, end_times):
                    lower_bounds, upper_bounds = [ np.r_[r, t]
                        for r, t in zip(bounds[i], (start_time, end_time)) ]
                    new_bounds.append((lower_bounds, upper_bounds))

    if noname:
        return regions['']
    else:
        return regions


def _enumerate(it):
    for ix, elem in enumerate(it):
        yield (ix,)+tuple(elem)

def group_roi(a, *args, overlap=.75, return_matches_only=False):
    """ Returns the smallest rectangles that contain the ROI grouped depending on whether
    they overlap or not.

    A ROI should be a pair (tuple) of `numpy.ndarray` for (lower-,upper-) bounds.
    
    Time is handled just like any space coordinate.
    As a consequence, if time bounds are also specified, the corresponding values
    should be scaled so that time can be artificially related to space in the volume
    calculation carried out in the estimatation of the amount of overlap.
    """
    if args:
        b = args[0]
        if args[1:]:
            raise ValueError('arguments `overlap` and `return_matches_only` should be keyworded')
    else:
        b = a
    intra_grouping = a is b
    if intra_grouping:
        already_grouped = set()
    grouped_roi = []
    for _a, _a_lb, _a_ub in _enumerate(a):
        if intra_grouping and _a in already_grouped:
            continue
        #
        _a_area = np.prod(_a_ub - _a_lb)
        #
        _matches = []
        for _b, _b_lb, _b_ub in _enumerate(b):
            if intra_grouping and _b == _a:
                continue
            #
            _lb, _ub = np.maximum(_a_lb, _b_lb), np.minimum(_a_ub, _b_ub)
            if np.all(_lb < _ub):
                _b_area = np.prod(_b_ub - _b_lb)
                _max_area = min(_a_area, _b_area)
                _intersect_area = np.prod(_ub - _lb)
                _overlap = _intersect_area / _max_area
                if overlap <= _overlap:
                    _matches.append((_b_lb, _b_ub))
                    if intra_grouping:
                        already_grouped.add(_b)
        _lb, _ub = _a_lb, _a_ub
        if _matches:
            # compute the minimum rectangle that contains the union
            while _matches:
                _other_lb, _other_ub = _matches.pop()
                _lb, _ub = np.minimum(_lb, _other_lb), np.maximum(_ub, _other_ub)
        elif return_matches_only:
            continue
        grouped_roi.append((_lb, _ub))
    return grouped_roi


def distance_to_roi_center(roi_obj, sampling):
    """
    Arguments:

        roi_obj (IndividualROI): ROI object

        sampling (Partition or Analysis): tessellation

    Returns:

        ndarray: distance between each cell center and the ROI center

    """
    a = roi_obj._eldest_parent # parent analyzer object
    roi_center = (roi_obj.bounding_box[0] + roi_obj.bounding_box[1]) / 2
    cell_centers = a.time.get_spatial_segmentation(sampling).cell_centers
    if cell_centers.shape[1] < roi_center.size:
        if cell_centers.shape[1] == roi_center.size - 1:
            roi_center = roi_center[:-1]
        else:
            raise ValueError('too many dimensions for roi boundaries')
    r = cell_centers - roi_center[np.newaxis,:]
    dist = np.sqrt(np.sum(r * r, axis=1))
    return dist


#def iter_maps


__all__ = [ 'set_contiguous_time_support_by_count', 'epanechnikov_density', 'density_based_roi',
        'group_roi', 'distance_to_roi_center' ]




# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from ..attribute import *
from ..artefact import *
from .abc import *
import os.path
import numpy as np


class ImagesInitializer(Initializer):
    """
    Initial value for the :class:`~tramway.analyzer.RWAnalyzer`
    :attr:`~tramway.analyzer.RWAnalyzer.images` attribute.

    *from_...* methods alters the parent attribute which specializes
    into an initialized :class:`Images` object.
    """
    __slots__ = ()
    def from_tiff_file(self, filepath):
        """
        Defines the raw single molecule imaging file.

        Loading is not performed while calling this method.
        """
        self.specialize( StandaloneTiffFile, filepath )
    def from_tiff_files(self, filepattern):
        """
        Defines the raw single molecule imaging files.

        Loading is not performed while calling this method.
        """
        self.specialize( TiffFiles, filepattern )


class ImageParameters(object):
    __slots__ = ()
    # default implementation for children classes that define `_pixel_size` and `_loc_offset`
    def __init__(self):
        self._pixel_size = self._loc_offset = None
    @property
    def pixel_size(self):
        r"""
        *float*: Pixel size in :math:`\mu m`
        """
        return self._pixel_size
    @pixel_size.setter
    def pixel_size(self, pxsize):
        self._pixel_size = pxsize
    @property
    def loc_offset(self):
        """
        *numpy.ndarray*: Offset between coordinates and the image, in pixels
        """
        return self._loc_offset
    @loc_offset.setter
    def loc_offset(self, offset):
        self._loc_offset = offset
    # access to shared parameters
    @property
    def frame_interval(self):
        """
        *float*: See :attr:`~tramway.analyzer.spt_data.SPTParameters.frame_interval`
        """
        return self._eldest_parent.spt_data.frame_interval
    @frame_interval.setter
    def frame_interval(self, dt):
        self._eldest_parent.spt_data.frame_interval = dt
    @property
    def dt(self):
        """
        *float*: See :attr:`~tramway.analyzer.spt_data.SPTParameters.dt`
        """
        return self.frame_interval
    @dt.setter
    def dt(self, dt):
        self.frame_interval = dt


class _RawImage(AnalyzerNode, ImageParameters):
    __slots__ = ('_stack', '_pixel_size', '_loc_offset')
    def __init__(self, stack, **kwargs):
        AnalyzerNode.__init__(self, **kwargs)
        ImageParameters.__init__(self)
        self._stack = stack
    @property
    def stack(self):
        if callable(self._stack):
            self._stack = self._stack()
        return self._stack
    @stack.setter
    def stack(self, stack):
        self._stack = stack
    @property
    def n_frames(self):
        return len(self.stack)
    @property
    def width(self):
        return first(self.stack).shape[1]
    @property
    def height(self):
        return first(self.stack).shape[0]

    def as_frames(self, index=None, return_time=False):
        """
        Generator function; iterates over the image frames and yields
        NumPy 2D arrays, or pairs of (*float*, NumPy 2D array).

        Arguments:

            index (*int*, *Set*, *Sequence* or *callable*):
                frame filter; see also :func:`~tramway.analyzer.attribute.indexer`.

            return_time (bool): return time along with image frames, as first item.

        """
        for f, frame in indexer(index, self.stack, return_index=True):
            if return_time:
                t = (f+1)*self.dt
                yield t, frame
            else:
                yield frame

    def cropping_bounds(self, bounding_box, relative=True):
        """
        Discretize bounding box spatial coordinates to clip them
        onto image pixels.

        Pixels are assumed to be centered on their indices in the array,
        when plotted.

        .. note::

            `cropping_bounds` is supposed to work well in combination with
            `crop_frames`.
            However, a notable exception is out-of-bounds indices.
            `crop_frames` adjusts the size of the pixel matrix wrt the image,
            while `cropping_bounds` does not min-max the returned indices wrt
            the natural bounds of the image.

        Arguments:

            bounding_box (tuple): pair of ndarrays (lower bound, upper bound).

            relative (bool): if ``True``, ignores the coordinate offset if any.

        Returns:

            tuple: pair of ndarrays; pixel-clipped lower bounds and upper bounds.

        """
        loc_offset = self.loc_offset
        if loc_offset is None or relative:
            loc_offset = 0.
        lb, ub = [ b / self.pixel_size - loc_offset for b in bounding_box ]
        lb = np.floor(lb) * self.pixel_size
        ub = np.ceil(ub) * self.pixel_size
        return lb, ub

    def crop_frames(self, bounding_box, index=None, return_time=False):
        """
        Generator function; iterates and crops the image frames, similarly to
        :meth:`as_frames`.

        .. note::

            Time bounds are not supported yet.

        Arguments:

            bounding_box (tuple): pair of NumPy arrays (lower bound, upper bound).

            index (*int*, *Set*, *Sequence* or *callable*):
                frame filter; see also :func:`~tramway.analyzer.attribute.indexer`.

            return_time (bool): return time along with cropped image frames, as first item.

        """
        lb, ub = bounding_box
        if 2<lb.size:
            self.logger.warning('time supports are not supported yet')
            lb, ub = lb[:2], ub[:2]
            bounding_box = (lb, ub)
        loc_offset = self.loc_offset
        if loc_offset is None:
            loc_offset = 0.
        lb, ub = [ b / self.pixel_size - loc_offset for b in bounding_box ]
        lb = np.floor(lb).astype(int)
        ub = np.ceil(ub).astype(int)
        if not np.all(lb <= ub):
            raise ValueError('image cropping failed: lower bound > upper bound')
        i_min = max(0, self.height-1-ub[1]) # range start (included)
        i_max = min(self.height-lb[1], self.height) # range stop (excluded)
        j_min = max(0, lb[0]) # range start (included)
        j_max = min(ub[0]+1, self.width) # range stop (excluded)
        if return_time:
            for t, frame in self.as_frames(index, return_time):
                yield t, frame[i_min:i_max, j_min:j_max]
        else:
            for frame in self.as_frames(index, return_time):
                yield frame[i_min:i_max, j_min:j_max]

    def to_color_movie(self, output_file=None, fourcc='VP80', colormap='gray',
            locations=None, trajectories=None, frames=None, origin=None,
            markersize=2, linecolor='many', linewidth=None,
            magnification=None, playback_rate=None, light_intensity=1.):
        """
        Generates a movie of the images with overlaid locations or trajectories.

        Arguments:

            output_file (str): path of the movie file.

            fourcc (str): 4-character code string.

            colormap (*str* or *matplotlib.colors.ListedColormap*): Matplotlib colormap;
                see also https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html.

            locations (pandas.DataFrame): particle locations with columns :const:`'x'`,
                :const:`'y'` and :const:`'t'`.

            trajectories (pandas.DataFrame): trajectories with columns :const:`'n'`,
                :const:`'x'`, :const:`'y'` and :const:`'t'`;
                note this differs from translocations in that displacements are not
                encoded along with locations and trajectory terminations are independent rows.

            frames (sequence): iterable of time (*float*, in seconds) and frame (2D pixel array)
                pairs; if `origin` defines time, `frames` can be a sequence of frames only;
                :meth:`as_frames` is called instead if undefined.

            origin (*numpy.ndarray* or *pandas.Series*): data lower bound if `frames` is defined;
                implicit columns are :const:`'x'` and :const:`'y'`;
                to define time, `origin` must be a `Series`
                with indices :const:`'x'`, :const:`'y'` and :const:`'t'`;
                `origin` is useful only for overlaying locations or trajectories.

            markersize (int): location marker size in pixels (square side).

            linecolor (*str* or 3-column float array): color for trajectories;
                value :const:`None` defaults to red.

            linewidth (*int* or *float*): trajectory line width.

            magnification (*int* or *str*): the original image pixels can be represented as
                square-patches of `magnification` video pixel side;
                if *str*:
                :const:`'1x'` = `round(pixel_size/localization_precision)`,
                :const:`'2x'` = `round(2*pixel_size/localization_precision)`;
                :const:`'2x'` is adequate for overlaid trajectories, even with the over-compressing
                :const:`'MJPG'` encoder.

            playback_rate (float): default playback rate;
                1 is real-time, 0.5 is half the normal speed.

            light_intensity (float): scale the pixel intensity values;
                works best with the 'gray' colormap.

        """
        import cv2
        # TODO: support PIL as an optional replacement for skimage
        from skimage.util import img_as_ubyte
        from . import draw
        from matplotlib import cm, colors

        if output_file is None:
            try:
                output_file = os.path.splitext(self.filepath)[0]+'.avi'
            except AttributeError:
                raise ValueError('output_file is not defined')

        dt = self.dt
        pxsize = self.pixel_size
        if pxsize is None:
            pxsize = 1.

        if self.loc_offset is None:
            offset = np.zeros((1,2), dtype=np.float)
        else:
            offset = np.asarray(self.loc_offset)
            if not offset.shape[1:]:
                offset = offset[np.newaxis,:]
        #x_offset, y_offset = offset[0,0], offset[0,1]

        if origin is None:
            xy0 = None
            t0 = None
        elif isinstance(origin, np.ndarray):
            xy0 = origin.reshape((1,2)).astype(np.float)
            t0 = None
        else:#if isinstance(origin, pd.Series):
            xy0 = origin[list('xy')].values[np.newaxis,:].astype(np.float)
            try:
                t0 = origin['t'].astype(np.float)
            except KeyError:
                t0 = None

        if magnification is None:
            magnification = 1
        elif isinstance(magnification, str) and magnification.endswith('x'):
            try:
                magnification = float(magnification[:-1])
            except ValueError:
                raise ValueError('cannot parse the magnification factor')
            try:
                magnification *= self.pixel_size / self._eldest_parent.localization_precision
            except (AttributeError, TypeError):
                raise ValueError('failed to adjust magnification; pixel_size or localization_precision not defined')
            magnification = int(np.round(magnification))

        if playback_rate is None:
            playback_rate = 1.

        if light_intensity == 1:
            color_scale = None
        else:
            color_scale = int(np.round(1./light_intensity))

        marker_color = line_color = np.array([[1,0,0]], dtype=np.float) # red
        if locations is not None:
            if callable(locations):
                locations = locations()
            if markersize % 2:
                marker_size_delta = (markersize - 1) / 2
                def _floor(x):
                    return np.round(x) - marker_size_delta
                def _ceil(x):
                    return np.round(x) + marker_size_delta
            else:
                marker_size_delta = markersize / 2 - 1
                def _floor(x):
                    return np.floor(x) - marker_size_delta
                def _ceil(x):
                    return np.floor(x) + 1 + marker_size_delta
        if trajectories is not None:
            if callable(trajectories):
                trajectories = trajectories()
            if isinstance(linecolor, str):
                if linecolor == 'many':
                    from tramway.plot.mesh import __colors__
                    line_color = np.stack([colors.to_rgb(c) for c in __colors__], axis=0)
                else:
                    line_color = colors.to_rgb(linecolor)[np.newaxis,:]
            elif isinstance(linecolor, np.ndarray):
                line_color = linecolor
            lc = line_color
            def append_n(n, ijk):
                i, j, k = ijk
                return i, j, k, np.full(k.shape, n)
            trajectories_as_dict = {}
            #trajectories_as_dict = { n: trajectories[list('xyt')][trajectories['n']==n]
            #        for n in np.unique(trajectories['n']) }
        dt_max_err = (.5*dt)**2
        def isclose(ts, t):
            _dt = ts-t
            return _dt*_dt < dt_max_err

        vid = None # will be initialized later
        vid_pxsize = pxsize / magnification
        mag_offset = .5 * (magnification - 1)
        vid_offset = offset / magnification - mag_offset

        if isinstance(colormap, str):
            colormap = cm.get_cmap(colormap)
        elif not callable(colormap):#isinstance(colormap, colors.ListedColormap):
            colormap = cm.viridis

        if frames:
            if not (callable(frames) or t0 is None):
                def wrap(fs, t0, dt):
                    for i, f in enumerate(fs):
                        if isinstance(f, (tuple, list)):
                            t, f = f
                            if i == 0 and t != t0:
                                raise ValueError('start time does not match with first frame time: {} != {}'.format(t0, t))
                        else:
                            t = t0 + i * dt
                        yield t, f
                frames = wrap(frames, t0, dt)
        else:
            frames = self.as_frames

        t_prev = -1
        for t, frame in frames(return_time=True) if callable(frames) else frames:
            if t <= t_prev:
                raise ValueError('time does not strictly increase')
            t_prev = t

            if vid is None:
                # complete the initialization
                height, width = frame.shape
                ii_max = height * magnification - 1
                jj_max = width * magnification - 1

                vid = cv2.VideoWriter(os.path.expanduser(output_file),
                        cv2.VideoWriter_fourcc(*fourcc),
                        playback_rate/dt, (width*magnification, height*magnification), True)

            if color_scale:
                frame = frame // color_scale

            frame = colormap(frame)[:,:,:3]

            if magnification:
                frame = np.repeat(np.repeat(frame, magnification, axis=0), magnification, axis=1)

            if trajectories is not None:
                # anti-aliasing
                lw = 0 if linewidth is None else linewidth
                aa = bool(linewidth)

                # get the ids of the active trajectories for frame f
                active_trajectories = trajectories['n'][ isclose(trajectories['t'], t) ]
                assert all_unique(active_trajectories)

                # extract the corresponding coordinate series truncated at time t
                trajs_f = []
                for n in active_trajectories:
                    try:
                        xyt_n = trajectories_as_dict[n]
                    except KeyError:
                        trajectories_as_dict[n] = xyt_n = trajectories[trajectories['n']==n]
                    xy_n = xyt_n[list('xy')][xyt_n['t']<t+.5*dt]
                    if len(xy_n)<2:
                        continue
                    if xy0 is None:
                        traj = xy_n.values / vid_pxsize - vid_offset
                    else:
                        traj = (xy_n.values - xy0) / vid_pxsize + mag_offset
                    if np.any(traj < 0): # this may occur with negative offsets
                        continue
                    if not aa:
                        traj = np.round(traj).astype(np.uint32)
                    trajs_f.append((n, traj))

                # overlay the truncated trajectories
                if trajs_f:
                    jj, ii, kk, nn = zip(*[ append_n(n,
                        draw.multiline_aa(traj[:,0], traj[:,1], thickness=lw)
                        ) for n, traj in trajs_f ])
                    ii = np.concatenate(ii)
                    jj = np.concatenate(jj)
                    ok = (0<=ii)&(ii<=ii_max)&(0<=jj)&(jj<=jj_max)
                    ii, jj = ii[ok], jj[ok]
                    kk = np.concatenate(kk)[ok][:,np.newaxis]
                    if 1<len(line_color):
                        nn = np.concatenate(nn)[ok]
                        lc = line_color[nn % len(line_color)]
                    frame[ii_max-ii,jj,:] = (1.-kk) * frame[ii_max-ii,jj,:] + kk * lc

            if locations is not None:
                xy = locations[ isclose(locations['t'], t) ]
                if xy0 is None:
                    xy_f = xy[list('xy')].values / vid_pxsize - vid_offset
                else:
                    xy_f = (xy[list('xy')].values - xy0) / vid_pxsize + mag_offset
                for j,i in xy_f:
                    i = np.array([_floor(i), _ceil(i)], dtype=np.int)
                    j = np.array([_floor(j), _ceil(j)], dtype=np.int)
                    if np.any(i<0) or np.any(j<0):
                        continue
                    i,j = np.meshgrid(
                            np.arange(i[0],i[1]+1),
                            np.arange(j[0],j[1]+1), indexing='ij')
                    try:
                        frame[ii_max-np.ravel(i),np.ravel(j),:] = marker_color
                    except IndexError:
                        pass

            vid.write(img_as_ubyte(frame)[:,:,::-1])
        vid.release()

    @property
    def _mpl_impl(self):
        from .mpl import Mpl
        return Mpl
    @property
    def mpl(self):
        """ tramway.analyzer.images.mpl.Mpl: Matplotlib utilities """
        return self._mpl_impl(self)


class RawImage(_RawImage):
    __slots__ = ()

Image.register(RawImage)

class _ImageFile(_RawImage):
    __slots__ = ('_filepath',)
    def __init__(self, filepath, **kwargs):
        self._filepath = filepath
        RawImage.__init__(self, self.read, **kwargs)
    @property
    def filepath(self):
        return self._filepath
    @filepath.setter
    def filepath(self, fp):
        self._filepath = fp
    def read(self):
        from skimage import io
        img = io.MultiImage(os.path.expanduser(self.filepath))
        if len(img) == 1 and 2<len(img[0].shape):
            img = img[0]
        return img

class ImageFile(_ImageFile):
    __slots__ = ()

Image.register(ImageFile)

class _TiffFile(_ImageFile):
    __slots__ = ()

class TiffFile(ImageFile):
    __slots__ = ()

class StandaloneImage(object):
    __slots__ = ()
    def __len__(self):
        return 1
    def __iter__(self):
        yield self

class ImageIterator(AnalyzerNode, ImageParameters):
    """ Partial :class:`Images` implementation for multiple SPT data items.

    Children classes must implement the :meth:`__iter__` method."""
    __slots__ = ()
    def __init__(self, **kwargs):
        AnalyzerNode.__init__(self, **kwargs)
        #ImageParameters.__init__(self)
    @property
    def pixel_size(self):
        it = iter(self)
        pxsize = next(it).pixel_size
        while True:
            try:
                _pxsize = next(it).pixel_size
            except StopIteration:
                break
            else:
                if _pxsize != pxsize:
                    raise AttributeError('not all the images share the same pixel size')
        return pxsize
    @pixel_size.setter
    def pixel_size(self, pxsize):
        for f in self:
            f.pixel_size = pxsize
    @property
    def loc_offset(self):
        it = iter(self)
        offset = next(it).loc_offset
        while True:
            try:
                _offset = next(it).loc_offset
            except StopIteration:
                break
            else:
                if not np.all(_offset == offset):
                    raise AttributeError('not all the images share the same localization offset')
        return offset
    @loc_offset.setter
    def loc_offset(self, offset):
        for f in self:
            f.loc_offset = offset

class StandaloneImageFile(_ImageFile, StandaloneImage):
    __slots__ = ()

Images.register(StandaloneImageFile)

class StandaloneTiffFile(_TiffFile, StandaloneImage):
    __slots__ = ()

Images.register(StandaloneTiffFile)


class RawImages(ImageIterator):
    __slots__ = ('_images',)
    def __init__(self, stacks, **kwargs):
        ImageIterator.__init__(self, **kwargs)
        self.images = stacks
    @property
    def images(self):
        return self._images
    @images.setter
    def images(self, stacks):
        self._images = tuple([ self._bear_child( RawImage, stack ) for stack in stacks ])
    @property
    def reified(self):
        return True
    def __len__(self):
        return len(self.images)
    def __iter__(self):
        yield from self.images

#Images.register(RawImages)

class ImageFiles(ImageIterator):
    __slots__ = ('_files', '_filepattern')
    def __init__(self, filepattern, **kwargs):
        ImageIterator.__init__(self, **kwargs)
        self._files = []
        if isinstance(filepattern, str):
            self._filepattern = os.path.expanduser(filepattern)
        else:
            self._filepattern = [os.path.expanduser(pattern) for pattern in filepattern]
    @property
    def filepattern(self):
        return self._filepattern
    @filepattern.setter
    def filepattern(self, fp):
        if self._files:
            if fp != self._filepattern:
                raise AttributeError('the files have already been listed; cannot set the file pattern anymore')
        else:
            self._filepattern = fp
    @property
    def files(self):
        if not self._files:
            self.list_files()
        return self._files
    @property
    def filepaths(self):
        return [ f.filepath for f in self.files ]
    @property
    def partially_reified(self):
        return self._files and any([ f.reified for f in self._files ])
    @property
    def fully_reified(self):
        return self._files and all([ f.reified for f in self._files ])
    @property
    def reified(self):
        return self.fully_reified
    def __len__(self):
        return len(self.files)
    def __iter__(self):
        yield from self.files
    def list_files(self, _element_cls=ImageFile):
        from glob import glob
        self._files = glob(self.filepattern)
        if not self._files:
            raise ValueError("no files found")
        self._files = [ self._bear_child( _element_cls, filepath ) for filepath in self._files ]

Images.register(ImageFiles)

class TiffFiles(ImageFiles):
    __slots__ = ()
    def list_files(self):
        ImageFiles.list_files(self, TiffFile)

def all_unique(values):
    return np.unique(values).size == values.size


__all__ = ['Images', 'Image', 'ImagesInitializer', 'ImageParameters', '_RawImage',
        'RawImage', 'ImageFile', 'TiffFile', 'StandaloneImageFile', 'StandaloneTiffFile',
        'RawImages', 'ImageFiles', 'TiffFiles']




# -*- coding: utf-8 -*-

# Copyright © 2020-2021, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from functools import wraps

from tramway.core.analyses import abc
#from tramway.core.analyses.base import Analyses
from tramway.core.analyses.lazy import Analyses

import pkg_resources
import platform
import time


def standard_metadata():
    return dict(
        os = platform.system(),
        python = platform.python_version(),
        tramway = pkg_resources.get_distribution('tramway').version,
        datetime = time.strftime('%Y-%m-%d %H:%M:%S UTC%z'),
        )

class Analysis(object):
    __slots__ = ('_parent', 'label', '_data', '_subtree')
    def __init__(self, data, parent=None):
        self._parent = parent
        self.label = None
        self._data = data
        self._subtree = None
    @property
    def parent(self):
        parent = self._parent
        if isinstance(parent, Analysis):
            parent = parent._subtree
        elif not isinstance(parent, abc.Analyses):
            try: # roi
                parent = parent._spt_data
            except AttributeError:
                pass
            parent = parent.analyses
        return parent
    @property
    def data(self):
        return self._data
    @data.setter
    def data(self, d):
        self._data = d
        self._parent = self._subtree
        self._subtree = None
        self.label = None
    @property
    def subtree(self):
        return self._subtree
    @property
    def is_input(self):
        return self._subtree is None
    @property
    def is_output(self):
        return not self.is_input
    def commit_as_analysis(self, label, comment=None):
        label = self.parent.autoindex(label)
        subtree = Analyses(self.data, standard_metadata())
        self.parent.add(subtree, label, comment)
        self._subtree = self.parent[label]
        return label
    @classmethod
    def get_analysis(cls, parent, label):
        artefact = cls(None, parent)
        if label is None:
            from tramway.analyzer.attribute import single
            try:
                label = single(artefact.parent.labels)
            except RuntimeError:
                raise RuntimeError('undefined label for analysis artefact') from None
        artefact.label = label
        artefact._subtree = artefact.parent[label]
        artefact._data = artefact.subtree.data
        return artefact
    @property
    def analyses(self):
        if self.is_input:
            return self.parent
        else:
            return self.subtree
    def get_child(self, label=None):
        return Analysis.get_analysis(self, label)
    def get_parent(self):
        if isinstance(self._parent, Analysis):
            return self._parent
        else:
            return None
    @classmethod
    def rebase_tree(cls, tree, origin_tree):
        """ Merges `tree` into pre-existing `origin_tree`.

        `tree` is modified in-place.
        As a consequence, the resulting tree is `tree`,
        and NOT `origin_tree`, although the operation
        is conceptually easier to describe as `tree` being
        merged into `origin_tree`, or equivalently
        `origin_tree` being updated with `tree`.

        `tree` takes priority and an artefact in `origin_tree`
        is taken into account only if the corresponding label does
        not already exist in `tree`.

        `rebase_tree` has been designed to rebase `tree` from
        `origin_tree` and pulling any change.
        """
        subtree0, subtree1 = tree, origin_tree
        if subtree0:
            labels0 = set(list(subtree0.labels))
            labels1 = set(list(subtree1.labels))
            for label in labels0 & labels1:
                cls.rebase_tree(subtree0[label], subtree1[label])
            for label in labels1 - labels0:
                subtree0[label] = subtree1[label]
                try:
                    subtree0.comments[label] = subtree1.comments[label]
                except KeyError:
                    pass
        else:
            if subtree1:
                for label in subtree1:
                    subtree0[label] = subtree1[label]
                    try:
                        subtree0.comments[label] = subtree1.comments[label]
                    except KeyError:
                        pass
    @classmethod
    def save(cls, filepath, tree, **kwargs):
        """
        Wrapper for :func:`~tramway.core.hdf5.store.save_rwa` that overrides
        the default values for `compress` (:const:`False`) and `append` (:const:`True`).

        `append=None` is interpreted as `append=False` and a `PendingDeprecationWarning`
        warning is thrown.

        `append` behavior itself is also overriden.
        It is replaced by that of :meth:`rebase_tree`.
        As a consequence, :meth:`save` also understands keyworded argument `rebase`,
        and `append` is treated as a synonym for `rebase`.
        """
        import os.path
        from tramway.core.hdf5.store import load_rwa, save_rwa
        kwargs['compress'] = kwargs.get('compress', False)
        append = kwargs.pop('append', None)
        rebase = kwargs.pop('rebase', None)
        if rebase is None:
            rebase = True if append is None else append
        elif not (append is None or append == rebase):
            raise ValueError('`append` and `rebase` do not have compatible values')
        if rebase and os.path.isfile(filepath) and os.path.getsize(filepath) > 0:
            existing_tree = load_rwa(filepath)
            cls.rebase_tree(tree, existing_tree)
        save_rwa(filepath, tree, **kwargs)


def _getter(pos, kw=None, default=None):
    if kw is None:
        def _get(_args, _kwargs):
            return _args[pos]
    else:
        def _get(_args, _kwargs):
            try:
                return _args[pos]
            except IndexError:
                return _kwargs.get(kw, default)
    return _get
def _setter(pos, kw=None):
    if kw is None:
        def _set(_args, _kwargs, val):
            return _args[:pos]+(val,)+_args[pos+1:]
    else:
        def _set(_args, _kwargs, val):
            try:
                _args[pos]
            except IndexError:
                _kwargs[kw] = val
            else:
                return _args[:pos]+(val,)+_args[pos+1:]
    return _set

def _wrap_method(met, _get, _set):
    @wraps(met)
    def wrapper_method(self, *args, **kwargs):
        _arg = _get(args, kwargs)
        wrap = isinstance(_arg, Analysis)
        if wrap:
            wrapper = _arg
            args = _set(args, kwargs, wrapper.data)
        res = met(self, *args, **kwargs)
        if wrap:
            res = type(wrapper)(res, wrapper.analyses)
        return res
    return wrapper_method

def _wrap(pos, kw=None):
    return lambda met: _wrap_method(met, _getter(pos, kw), _setter(pos, kw))

def analysis(met_or_pos=None, kw=None):
    if met_or_pos is None:
        if kw is not None:
            raise ValueError('keyword is defined whereas argument position is not')
        return _wrap(0)
    elif isinstance(met_or_pos, int):
        pos = met_or_pos - 1
        if kw is None:
            return _wrap(pos)
        elif isinstance(kw, str):
            return _wrap(pos, kw)
        else:
            raise TypeError('keyword is not a string')
    elif callable(met_or_pos):
        return _wrap_method(met_or_pos, _getter(0), _setter(0))
    else:
        raise TypeError('argument index is not an integer')



def commit_as_analysis(label, analysis, parent=None):
    if not isinstance(analysis, Analysis):
        if parent is None:
            raise ValueError('could not find the analysis tree')
        analysis = Analysis(analysis, parent)
    analysis.commit_as_analysis(label)
    return analysis


__all__ = ['Analysis', 'analysis', 'commit_as_analysis', 'standard_metadata']




# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from . import abc
from . import base
from . import lazy
import warnings
import os.path


class EventDict(dict):
    """ dictionnary that can trigger an event on `__setitem__`. """
    def __init__(self, d, preprocess, postprocess):
        dict.__init__(self, d)
        self.preprocess = preprocess
        self.postprocess = postprocess
    def __setitem__(self, k, v):
        dict.__setitem__(self, k, self.preprocess(v))
        self.postprocess(self, k)


class WithState(object):
    """ complementary class for flagging an analysis as modified.
    
    Any slotted child class should define attribute `_modified` """
    __slots__ = ()
    def __init__(self, *args, cls=base.Analyses):
        cls.__init__(self, *args)
        self._modified = False
    def flag_as_modified(self):
        self._modified = True
    def reset_modification_flag(self, recursive=False):
        self._modified = False
        if recursive:
            for child in self.instances.values():
                child.reset_modification_flag(True)
    def modified(self, recursive=False):
        if self._modified:
            return True
        elif recursive:
            return any( child.modified(True) for child in self.instances.values() )


def with_state(cls, handler):
    """ derivate class `cls` so that it also inherits from `WithState`.

    If `cls` is a class instance instead, then it copies `cls` into an object
    that inhertits from `type(cls)` and `WithState`.

    The resulting class or class instance features a `statefree` method that
    returns a copy of `self` in the original parent type.
    """
    if isinstance(cls, type):
        assert cls in (base.Analyses, lazy.Analyses)
        class AnalysesWithState(cls, WithState):
            __slots__ = ('_modified',)
            def __init__(self, data=None, metadata=None):
                WithState.__init__(self, data, metadata, cls=base.Analyses)
                self._instances = EventDict(self._instances, handler.preprocess(self), handler.postprocess(self))
            def statefree(self):
                analyses = cls(self.data, self.metadata)
                analyses._comments = self._comments
                analyses._instances = { label: self[label].statefree() for label in self.labels }
                return analyses
        return AnalysesWithState
    else:
        analyses = cls
        if isinstance(analyses, Analyses):
            analyses = analyses.analyses
            assert not isinstance(analyses, Analyses)
            return analyses
        cls = with_state(type(analyses), handler)
        def _with_state(_analyses):
            assert isinstance(_analyses, (base.Analyses, lazy.Analyses))
            _state = cls(_analyses._data, _analyses._metadata)
            _state._comments = _analyses._comments
            _state._instances = EventDict(
                    { _label: _with_state(_analyses[_label]) for _label in _analyses.labels },
                    handler.preprocess(_state), handler.postprocess(_state),
                    )
            return _state
        return _with_state(analyses)


class AnalysesProxy(object):
    """ proxy for analyses objects """
    __slots__ = ('_analyses',)
    def __init__(self, data=None, metadata=None, cls=base.Analyses):
        if isinstance(data, cls):
            self.analyses = data
            if metadata:
                warnings.warn('ignoring argument `metadata`')
        else:
            self.analyses = cls(data, metadata)
    @property
    def analyses(self):
        return self._analyses
    @analyses.setter
    def analyses(self, a):
        self._analyses = a
    @property
    def data(self):
        return self.analyses.data
    @data.setter
    def data(self, d):
        self.analyses.data = d
    @property
    def _data(self):
        return self.analyses._data
    @_data.setter
    def _data(self, d):
        self.analyses._data = d
    @property
    def artefact(self):
        return self.analyses.artefact
    @artefact.setter
    def artefact(self, a):
        self.analyses.artefact = a
    @property
    def metadata(self):
        return self.analyses.metadata
    @metadata.setter
    def metadata(self, d):
        self.analyses.metadata = d
    @property
    def instances(self):
        return self.analyses.instances
    @property
    def _instances(self):
        return self.analyses._instances
    @property
    def comments(self):
        return self.analyses.comments
    @property
    def _comments(self):
        return self.analyses._comments
    @property
    def labels(self):
        return self.analyses.labels
    def keys(self):
        return self.analyses.keys()
    def autoindex(self, pattern=None):
        return self.analyses.autoindex(pattern)
    def add(self, analysis, label=None, comment=None, raw=False):
        self.analyses.add(analysis, label, comment, raw)
    def __nonzero__(self):
        return self.analyses.__nonzero__()
    def __len__(self):
        return self.analyses.__len__()
    def __missing__(self, label):
        return self.analyses.__missing__(label)
    def __iter__(self):
        return self.analyses.__iter__()
    def __contains__(self, label):
        return self.analyses.__contains__(label)
    def __getitem__(self, label):
        return self.analyses.__getitem__(label)
    def __setitem__(self, label, analysis):
        self.analyses.__setitem__(label, analysis)
    def __delitem__(self, label):
        self.analyses.__delitem__(label)
    def __str__(self):
        return self.analyses.__str__()

abc.Analyses.register(AnalysesProxy)

class LazyAnalysesProxy(AnalysesProxy):
    __slots__ = ()
    def __init__(self, data=None, metadata=None, cls=lazy.Analyses):
        AnalysesProxy.__init__(self, data, metadata, cls)
    @property
    def type(self):
        return self.analyses.type
    def terminate(self):
        self.analyses.terminate()


__autosavecapable_slots__ = ('_default_autosave_policy','_active_autosave_policy')
class AutosaveCapable(object):
    """
    Abstract class.

    Children classes, if slotted, should define:
    
    * `_default_autosave_policy` (bool or str),
    * `_active_autosave_policy` (bool or str),

    An autosave policy can take various values if of ``str`` type:

    * *'on completion'* saves once on normal completion of the entire process (default)
    * *'on termination'* saves once on termination of the entire process, whether it is successful or not
    * *'on every step'* saves on every successful step

    If `_active_autosave_policy` is ``True``, then `_default_autosave_policy` applies.

    """
    __slots__ = ()
    def __init__(self, autosave=True):
        """
        Arguments:

            autosave (bool or str): default autosaving policy

        """
        self._default_autosave_policy = autosave
        self._active_autosave_policy = None
    @property
    def autosave(self):
        if self._active_autosave_policy is None or self._active_autosave_policy is True:
            return bool(self._default_autosave_policy)
        else:
            return bool(self._active_autosave_policy)
    @autosave.setter
    def autosave(self, flag):
        if not (isinstance(flag, bool) or isinstance(flag, str)):
            raise TypeError('autosave supports only boolean and string values')
        self._default_autosave_policy = flag
    @property
    def autosave_policy(self):
        if isinstance(self._active_autosave_policy, bool):
            if self._default_autosave_policy is False:
                return None
            elif self._default_autosave_policy is True:
                return 'on completion'
            else:
                return self._default_autosave_policy
        else:
            return self._active_autosave_policy
    @property
    def active(self):
        return bool(self._active_autosave_policy)
    @property
    def save_on_completion(self):
        policy = self.autosave_policy
        return policy and \
                (policy.endswith('completion') or policy.endswith('termination'))
    @property
    def save_on_every_step(self):
        policy = self.autosave_policy
        return policy and policy.endswith('step')
    @property
    def force_save(self):
        policy = self.autosave_policy
        return policy and policy.endswith('termination')
    def __enter__(self):
        self.reset_modification_flag(True)
        return self
    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self.modified(True):
            self.reset_modification_flag(True)
            if self.force_save or (exc_type is None and self.save_on_completion):
                self.save()
        self._active_autosave_policy = None
    def autosaving(self, policy=None):
        if policy is None and self.autosave:
            policy = True
        self._active_autosave_policy = policy
        return self
    def save(self):
        """ should call `self.reset_modification_flag(True)` """
        raise NotImplementedError('abstract method')


class Analyses(LazyAnalysesProxy, AutosaveCapable):
    """ autosaving analyses. 
    
    Argument and attribute `rwa_file` designate the output file."""
    __slots__ = __autosavecapable_slots__ + ('rwa_file','save_options','hooks')
    def __init__(self, data=None, metadata=None, rwa_file=None, autosave=False):
        if isinstance(data, AutosaveCapable):
            raise TypeError('nested autosave-capable objects')
        LazyAnalysesProxy.__init__(self, data, metadata)
        AutosaveCapable.__init__(self, autosave)
        self.rwa_file = rwa_file
        self.save_options = dict(force=True, compress=False)
        self.hooks = []
    @property
    def analyses(self):
        return self._analyses
    @analyses.setter
    def analyses(self, a):
        self._analyses = with_state(a, self.handler)
    def add(self, analysis, label=None, comment=None, raw=False):
        if not (raw or isinstance(analysis, abc.Analyses)):
            analysis = lazy.Analyses(analysis)
        if isinstance(analysis, abc.Analyses) and not isinstance(analysis, WithState):
            analysis = with_state(analysis, self.handler)
        self.analyses.add(analysis, label, comment, raw)
    @classmethod
    def from_rwa_file(cls, input_file, output_file=None, **kwargs):
        from tramway.core.hdf5.store import load_rwa
        if output_file is None:
            output_file = input_file
        analyses = load_rwa(input_file, lazy=True,
                force_load_spt_data=False)
        analyses = cls(analyses, rwa_file=output_file, **kwargs)
    def save(self, out_of_context=False):
        if not (out_of_context or self.active):
            raise RuntimeError("method 'save' called from outside the context")
        for hook in self.hooks:
            hook(self.analyses)
        if self.rwa_file:
            from tramway.core.hdf5.store import save_rwa
            save_rwa(os.path.expanduser(self.rwa_file), self.analyses.statefree(), **self.save_options)
            self.analyses.reset_modification_flag(True)
        else:
            warnings.warn('no output file defined', RuntimeWarning)
    # WithState proxy
    def flag_as_modified(self):
        self.analyses.flag_as_modified()
    def modified(self, recursive=False):
        return self.analyses.modified(recursive)
    def reset_modification_flag(self, recursive=False):
        self.analyses.reset_modification_flag(recursive)
    # handler methods; not for direct use
    def preprocess(self, analyses):
        def _preprocess(analysis):
            if isinstance(analysis, abc.Analyses) and not isinstance(analysis, WithState):
                analysis = with_state(analysis, self.handler)
            return analysis
        return _preprocess
    def postprocess(self, analyses):
        def _postprocess(*args):
            if self.active:
                if self.save_on_every_step:
                    self.save()
                else:
                    analyses.flag_as_modified()
        return _postprocess
    @property
    def handler(self):
        class PP(object):
            def __init__(_self):
                _self.preprocess = self.preprocess
                _self.postprocess = self.postprocess
        return PP()
    def statefree(self):
        return self.analyses.statefree()

__all__ = ['Analyses', 'AutosaveCapable']




# -*- coding: utf-8 -*-

# Copyright © 2018, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from .base import *
from collections import OrderedDict
import numpy as np
import pandas as pd


setup = {
    'make_arguments': OrderedDict((
        ('cell_count', dict(type=int, help='number of cells')),
        ('lower_bound', dict(type=float, nargs='+', help='space-separated lower coordinates: x_min y_min ...')),
        ('upper_bound', dict(type=float, nargs='+', help='space-separated upper coordinates: x_max y_max ...')),
        ('avg_probability', ()),
        # avg_location_count allows to control avg_probability from the commandline
        ('avg_location_count', dict(args=('-c', '--location-count'), kwargs=dict(type=int, default=80, help='average number of locations per cell'), translate=True)),
        )),
    }


class RandomMesh(Voronoi):

    __slots__ = ('lower_bound', 'upper_bound', 'avg_probability')

    def __init__(self, scaler=None, avg_probability=None, lower_bound=None, upper_bound=None,
        cell_count=None, **kwargs):
        Voronoi.__init__(self, scaler)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        if cell_count is not None:
            avg_probability = 1. / float(cell_count)
        self.avg_probability = avg_probability

    def _preprocess(self, points):
        Voronoi._preprocess(self, points) # initialize `scaler`
        return points # ... but do not scale

    # cell_centers property
    @property
    def cell_centers(self):
        return self._cell_centers # no scaling

    @cell_centers.setter
    def cell_centers(self, centers):
        self._cell_centers = centers

    def tessellate(self, points, allow_empty_cells=False, **kwargs):
        points = self._preprocess(points)
        # find the bounding box
        if self.lower_bound is None:
            self.lower_bound = points.min(axis=0)
        elif isinstance(points, pd.DataFrame) and not isinstance(self.lower_bound, pd.Series):
            self.lower_bound = pd.Series(self.lower_bound, index=points.columns)
        if self.upper_bound is None:
            self.upper_bound = points.max(axis=0)
        elif isinstance(points, pd.DataFrame) and not isinstance(self.upper_bound, pd.Series):
            self.upper_bound = pd.Series(self.upper_bound, index=points.columns)
        # pick some centroids within the bounding box
        space_columns = self.scaler.columns
        n_centroids = int(1. / self.avg_probability)
        centroids = pd.DataFrame(np.random.rand(n_centroids, len(space_columns)), columns=space_columns)
        centroids *= self.upper_bound - self.lower_bound
        centroids += self.lower_bound
        # make the Voronoi tessellation
        Voronoi.tessellate(self, centroids)
        #
        if not allow_empty_cells:
            import copy
            cells = Partition(points, copy.deepcopy(self)) # copying new in 0.6
            #cells.cell_index = cells.cell_index(centroids, **kwargs)
            ok = 0 < cells.location_count
            #self.cell_centers = None # before 0.6; was not enough to reset
            Voronoi.tessellate(self, centroids[ok])


__all__ = ['setup', 'RandomMesh']

import sys
if sys.version_info[0] < 3:

    import rwa
    import tramway.core.hdf5 as rules
    random_mesh_exposes = rules.voronoi_exposes + list(RandomMesh.__slots__)
    rwa.hdf5_storable(
        rwa.default_storable(RandomMesh, exposes=random_mesh_exposes),
        agnostic=True)

    __all__.append('random_mesh_exposes')




# -*- coding: utf-8 -*-

# Copyright © 2020-2021, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


import os.path
from tramway.core.rc import __user_interaction__


class Client(object):
    """
    encapsulates the low-level API of Paramiko,
    and merely simplifies authentification.
    """
    __slots__ = ('_host','_conn','_sftp_client','_password','_options')
    def __init__(self, host=None, **options):
        self._host = host
        self._conn = None
        self._sftp_client = None
        self._password = None
        self._options = options
    @property
    def host(self):
        """
        account information in the following format: *username@hostname*
        """
        return self._host
    @host.setter
    def host(self, addr):
        self._host = addr
    @property
    def connection(self):
        if self._conn is None:
            self.connect()
        return self._conn
    @property
    def sftp_client(self):
        if self._sftp_client is None:
            self._sftp_client = self.connection.open_sftp()
        return self._sftp_client
    @property
    def password(self):
        """
        read-only property that requests a password from the user if a password is required.

        set attribute `_password` to explicitly pass a password.
        Note however plain text passwords are security breaches.
        """
        if self._password is None:
            if __user_interaction__ is True:
                import getpass
                self._password = getpass.getpass(self.host+"'s password: ")
            else:
                raise RuntimeError('a password is required')
        return self._password
    @property
    def options(self):
        """
        keyword arguments passed to `paramiko.client.SSHClient.connect`.

        additional supported options are:

        * *allow_password* (bool): makes `connect` crash if a password is required
                and *allow_password* is ``False``

        """
        return self._options
    def connect(self):
        try:
            import paramiko
        except ImportError:
            raise ImportError('package paramiko is required')
        try:
            user, host = self.host.split('@')
        except ValueError:
            user, host = None, self.host
        self._conn = paramiko.SSHClient()
        self._conn.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self._conn.load_system_host_keys()
        allow_password = self.options.pop('allow_password', None)
        try:
            self._conn.connect(host, 22, user, **self.options)
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            if allow_password is False:
                raise
            else:
                self._conn.connect(host, 22, user, self.password, **self.options)
        if allow_password is not None:
            self.options['allow_password'] = allow_password
    def exec(self, cmd, shell=False, logger=None):
        """
        runs command *cmd* on the remote host.

        if some executable is reported missing, set ``shell=True``.
        """
        if isinstance(cmd, (tuple, list)):
            cmd = ' '.join(cmd)
        if shell:
            cmd = 'bash -l -c "{}"'.format(cmd.replace('"',r'\"'))
        if logger is not None:
            logger.debug('running command: '+cmd)
        _in, _out, _err = self.connection.exec_command(cmd)
        out = _out.read()
        if out and not isinstance(out, str):
            out = out.decode('utf-8')
        err = _err.read()
        if err and not isinstance(err, str):
            err = err.decode('utf-8')
        return out, err
    def put(self, src, dest=None, confirm=False):
        """
        uploads *src* to remote file *dest*.

        see also `paramiko.sftp_client.SFTPClient.put`.
        """
        if dest is None:
            dest = src
        return self.sftp_client.put(src, dest, confirm=confirm)
    def get(self, target, dest=None):
        """
        downloads *target* as local file *dest*.

        see also `paramiko.sftp_client.SFTPClient.put`.
        """
        if dest is None:
            dest = target
        if target.startswith('~/'):
            target = target[2:]
        dest = os.path.expanduser(dest)
        self.sftp_client.get(target, dest)
    def download_if_missing(self, target, target_url, logger=None):
        """
        downloads *target_url* if *target* is missing on the remote host.

        note: command *wget* is run from the remote host.
        """
        try:
            info = self.sftp_client.lstat(target)
        except:
            info = None
        if info is None:
            if target_url is None:
                return True
            if logger is not None:
                logger.info('downloading {}...'.format(target_url))
            out, err = self.exec('wget '+target_url)
            if err and logger is not None:
                logger.error(err)
            return not err
    def close(self):
        if self._conn is not None:
            self._conn.close()


__all__ = ['Client']




# -*- coding:utf-8 -*-

# Copyright © 2017-2019, Institut Pasteur
#    Contributor: Maxime Duval

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""
Here are defined useful functions for training the unsupervised model.
"""

import numpy as np
import pandas as pd
import sklearn
import torch
import torch.utils.data
from torch import nn, optim
from torch.nn import functional as F
import tqdm


class TabDataset(torch.utils.data.Dataset):
    """torch Dataset subclass. Performs 0 mean and 1 variance on features.
    """

    def __init__(self, df, c_drop={}):
        self.Features = list(set(df.columns).difference(c_drop))
        self.df = df.loc[:, self.Features].dropna().sort_index(axis=1)
        self.dict_iloc_to_traj = dict(
            list(zip(np.arange(len(self.df)), self.df.index)))
        self.scaler = sklearn.preprocessing.StandardScaler()
        self.scaler.fit(self.df)

    def __len__(self):
        return self.df.shape[0]

    def __getitem__(self, index):
        return torch.tensor(self.scaler.transform(
            (self.df.iloc[index].values)
            .reshape(1, -1))).float()


class VAE(nn.Module):
    """
    Neural network as seen in Kingma, D. P., & Welling, M. (2013).
    Auto-encoding variational bayes. arXiv preprint arXiv:1312.6114.

    Parameters
    ----------
    input_dim : int, dimension of vectors we want to compress.
    hidden_dims : list of int, dimension(s) of the hidden layers of the neural
        network.
    latent_dim : int, dimension of the latent space on which we compress input
        vectors.
    ps : list of float between 0 and 1, dropout rates at each layer. Should
        have the same dimension as hidden_dims.
    """

    def __init__(self, input_dim, hidden_dims, latent_dim, ps):
        super(VAE, self).__init__()
        assert len(hidden_dims) == len(ps), 'dropouts and hidden dims must'
        'have same dimension'
        self.input_dim = input_dim
        self.hidden_dims = hidden_dims
        range_h = range(len(self.hidden_dims)-1)
        self.latent_dim = latent_dim
        self.ps = ps

        self.fc_in = nn.Linear(self.input_dim, self.hidden_dims[0])
        lin_enc = [nn.Linear(self.hidden_dims[i], self.hidden_dims[i+1])
                   for i in range_h]
        enc_layers = []
        for i in range(len(lin_enc)):
            enc_layers += [lin_enc[i], nn.ReLU(), nn.Dropout(p=self.ps[i+1])]
        self.fc_enc = nn.Sequential(*enc_layers)

        self.fc_latin1 = nn.Linear(self.hidden_dims[-1], self.latent_dim)
        self.fc_latin2 = nn.Linear(self.hidden_dims[-1], self.latent_dim)
        self.fc_latout = nn.Linear(self.latent_dim, self.hidden_dims[-1])

        lin_dec = [nn.Linear(self.hidden_dims[i+1], self.hidden_dims[i])
                   for i in reversed(range_h)]
        dec_layers = []
        for i in range(len(lin_enc)):
            dec_layers += [lin_dec[i], nn.ReLU(), nn.Dropout(p=self.ps[-i-2])]
        self.fc_dec = nn.Sequential(*dec_layers)

        self.fc_out = nn.Linear(self.hidden_dims[0], self.input_dim)

    def encode(self, x):
        h1 = F.dropout(F.relu(self.fc_in(x)), p=self.ps[0])
        h1 = self.fc_enc(h1)
        return self.fc_latin1(h1), self.fc_latin2(h1)

    def reparameterize(self, mu, logvar):
        std = torch.exp(0.5*logvar)
        eps = torch.randn_like(std)
        return eps.mul(std).add_(mu)

    def decode(self, z):
        h3 = F.dropout(F.relu(self.fc_latout(z)), p=self.ps[-1])
        h3 = self.fc_dec(h3)
        # return torch.sigmoid(self.fc_out(h3))
        return self.fc_out(h3)

    def forward(self, x):
        mu, logvar = self.encode(x)
        z = self.reparameterize(mu, logvar)
        return self.decode(z), mu, logvar


def loss_function_vae(beta=1):
    def tmp_func(recon_x, x, mu, logvar):
        # BCE = F.binary_cross_entropy(recon_x, x, reduction='sum')
        BCE = F.mse_loss(recon_x, x, reduction='sum')
        KLD = - beta * 0.5 * torch.sum(1 + logvar - mu.pow(2) - logvar.exp())
        return BCE + KLD, BCE, KLD
    return tmp_func


def train_vae(model, optimizer, loss_fct, data_loader, device,
              nb_epochs=3, disp_per_epoch=10):
    for epoch in tqdm.tqdm_notebook(range(nb_epochs)):
        model.train()
        train_loss = 0
        avg_loss = 0
        BCE_loss = 0
        KLD_loss = 0
        for batch_idx, data in tqdm.tqdm_notebook(enumerate(data_loader),
                                                  total=len(data_loader)):
            data = data.to(device)
            optimizer.zero_grad()
            recon_batch, mu, logvar = model(data)
            loss, BCE, KLD = loss_fct(recon_batch, data, mu, logvar)
            loss.backward()
            train_loss += loss.item()
            BCE_loss += BCE
            KLD_loss += KLD
            optimizer.step()
            if (batch_idx % int(len(data_loader) / disp_per_epoch) ==
                    int(len(data_loader) / disp_per_epoch) - 1):
                print('Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                    epoch+1, batch_idx * len(data), len(data_loader.dataset),
                    100. * batch_idx / len(data_loader),
                    train_loss / int(len(data_loader) / disp_per_epoch)))
                print('Mean BCE : {0:.6f}, Mean KLD : {1:.6f}'.format(
                    (BCE_loss / int(len(data_loader) / disp_per_epoch)),
                    (KLD_loss / int(len(data_loader) / disp_per_epoch))))
                avg_loss += train_loss
                train_loss = 0
                BCE_loss = 0
                KLD_loss = 0
        print('====> Epoch: {} Average loss: {:.4f}'.format(
            epoch+1, avg_loss / len(data_loader)))


def get_mu_logvar(dl, model, device):
    mus = []
    logvars = []
    for batch_idx, data in tqdm.tqdm_notebook(enumerate(dl), total=len(dl)):
        data = data.to(device)
        mu, logvar = model.encode(data)
        mus.append(mu.cpu().detach().numpy()[:, 0, :])
        logvars.append(logvar.cpu().detach().numpy()[:, 0, :])
    mus = np.concatenate(mus)
    logvars = np.concatenate(logvars)
    return mus, logvars
